#include "SH_MEMORYLCD.h"

//ショートカット
#define SCK_LOW digitalWrite(pinSCK, LOW)
#define SCK_HIGH digitalWrite(pinSCK, HIGH)
#define CS_LOW digitalWrite(pinSCS, LOW)
#define CS_HIGH digitalWrite(pinSCS, HIGH)
#define WAIT_SPI delayMicroseconds(SH_MLCD_MONO_SPIWAIT)
#define WAIT_1US delayMicroseconds(1)
#define WAIT_3US delayMicroseconds(3)

//フォントファイル
#include "MPLUS12.dat"

//コンストラクタ ====================================================
MLCD_MONO::MLCD_MONO(unsigned char vSCK, unsigned char vSI, unsigned char vCS, unsigned char vCOM, unsigned char vDISP){
	//縦横、1行あたりのバイト数を保存
	_width = SH_MLCD_MONO_PXWIDTH;
	_height = SH_MLCD_MONO_PXHEIGHT;
	_bytes = _width >> 3; //_width / 8;
	
	//画素状態記憶用配列を生成((幅 / 8bit) * 高さ)
	//最大12KByte(400 / 8 * 240 = 12,000)
	//配列は0xFFでクリアする
	short siz = _bytes * _height;
	R = new char[siz];
	memset(R, -1, siz);
	
	//使用ピン保存
	pinSCK = vSCK;
	pinSI = vSI;
	pinSCS = vCS;
	pinCOM = (digitalPinToPort(vCOM) == NOT_A_PIN)? 0xFF : vCOM;
	pinDISP = (digitalPinToPort(vDISP) == NOT_A_PIN)? 0xFF : vDISP;
	
	//外部COM、外部DISPを使用するかどうか
	useEcom = (pinCOM != 0xFF);
	useEdisp = (pinDISP != 0xFF);
	
	//COMフラグ初期化
	comFlag = false;
	
	//命令送信フラグ初期化
	busySPI = false;
}

//コンストラクタ(オーバーロード) ====================================
MLCD_MONO::MLCD_MONO(unsigned char vSCK, unsigned char vSI, unsigned char vCS){
	//縦横、ビットを保存
	_width = SH_MLCD_MONO_PXWIDTH;
	_height = SH_MLCD_MONO_PXHEIGHT;
	_bytes = _width >> 3; //_width / 8;
	
	//画素状態記憶用配列を生成((幅 / 8bit) * 高さ)
	//最大12KByte(400 / 8 * 240 = 12,000)
	//配列は0xFFでクリアする
	short siz = _bytes * _height;
	R = new char[siz];
	memset(R, -1, siz);
	
	//使用ピン保存
	pinSCK = vSCK;
	pinSI = vSI;
	pinSCS = vCS;
	pinCOM = 0xFF;
	pinDISP = 0xFF;
	
	//外部COM、外部DISPを使用するかどうか
	useEcom = false;
	useEdisp = false;
	
	//COMフラグ初期化
	comFlag = false;
	
	//命令送信フラグ初期化
	busySPI = false;
}

//デストラクタ ======================================================
MLCD_MONO::~MLCD_MONO(){
	//画素状態記憶用配列を削除
	delete[] R;
}

//SPIモドキのコマンド送出 ===========================================
//＠モード決定
void MLCD_MONO::send_mode(unsigned char val){
	//モード決定はMSBファーストでsend_dataと同一
	send_data(val);
}
//＠ゲートアドレス選択
void MLCD_MONO::send_addr(unsigned char val){
	//アドレスはLSBファースト(ビットの最下位から送る)
	for(char i=0; i<8; i++){
		digitalWrite(pinSI, (val>>i & 0x01));
		SCK_HIGH;
		WAIT_SPI;
		SCK_LOW;
	}
}
//＠データ送出
void MLCD_MONO::send_data(unsigned char val){
	//データはMSBファースト(ビットの最上位から送る)
	for(char i=0; i<8; i++){
		digitalWrite(pinSI, (val<<i & 0x80)>>7);
		SCK_HIGH;
		WAIT_SPI;
		SCK_LOW;
	}
}

//ファイルの部分読み取り ============================================
//＠ファイルの部分読み取り(2byte)
short MLCD_MONO::fr2(File srcf){
	unsigned char rtn[2];
	rtn[0] = srcf.read();
	rtn[1] = srcf.read();
	return (rtn[1] << 8 | rtn[0]);
}
//＠ファイルの部分読み取り(4byte)
long MLCD_MONO::fr4(File srcf){
	unsigned char rtn[4];
	rtn[0] = srcf.read();
	rtn[1] = srcf.read();
	rtn[2] = srcf.read();
	rtn[3] = srcf.read();
	return (rtn[3] << 24 | rtn[2] << 16 | rtn[1] << 8 | rtn[0]);
}

//初期化 ============================================================
void MLCD_MONO::init(bool initPIN){
	//ピンを初期化
	if(initPIN){
		pinMode(pinSCK, OUTPUT);
		pinMode(pinSI, OUTPUT);
		pinMode(pinSCS, OUTPUT);
		if(useEcom) pinMode(pinCOM, OUTPUT);
		if(useEdisp) pinMode(pinDISP, OUTPUT);
	}
	
	//DISP以外のすべての信号をLOWに
	digitalWrite(pinSCK, LOW);
	digitalWrite(pinSI, LOW);
	digitalWrite(pinSCS, LOW);
	if(useEcom) digitalWrite(pinCOM, LOW);
	
	//表示オン
	dispON();
	
	delay(2); //メモリ液晶初期化待ち時間
	
	//表示クリア
	clear();
}

//表示クリア ========================================================
void MLCD_MONO::clear(){
	//コマンド送出フラグ
	busySPI = true;
	
	CS_HIGH;
	delay(25);
	send_mode(SH_MLCD_MONO_CLEAR | ((comFlag)? SH_MLCD_MONO_COM : SH_MLCD_MONO_DUMMY));
	send_data(SH_MLCD_MONO_DUMMY);
	delay(25);
	CS_LOW;
	WAIT_1US;
	
	//コマンド送出フラグ
	busySPI = false;
}

void MLCD_MONO::cls(){ //void clear()と同一
	clear();
}

//内部配列クリア(0xFFで初期化) ======================================
void MLCD_MONO::clearArray(){
	short siz = _bytes * _height;
	memset(R, -1, siz);
}

void MLCD_MONO::cla(){ //void clearArray()と同一
	clearArray();
}

//表示オン、オフ ====================================================
//＠表示オン
void MLCD_MONO::dispON(){
	if(useEdisp) digitalWrite(pinDISP, HIGH);
}
//＠表示オフ
void MLCD_MONO::dispOFF(){
	if(useEdisp) digitalWrite(pinDISP, LOW);
}

//COM反転 ===========================================================
void MLCD_MONO::invertCOM(){
	//コマンド送出中であればフラグを反転させるだけ
	//コマンド送出がない場合はCSを切り替えて強制的にCOM反転する
	if(busySPI){
		if(useEcom){
			digitalWrite(pinCOM, HIGH);
			WAIT_1US;
			digitalWrite(pinCOM, LOW);
		}
	}else{
		CS_HIGH;
		WAIT_1US;
		if(useEcom){
			digitalWrite(pinCOM, HIGH);
			WAIT_1US;
			digitalWrite(pinCOM, LOW);
		}else{
			send_mode((comFlag)? SH_MLCD_MONO_DUMMY : SH_MLCD_MONO_COM);
			send_data(SH_MLCD_MONO_DUMMY);
		}
		WAIT_3US;
		CS_LOW;
		WAIT_1US;
	}
	comFlag = !comFlag;
}

//配列データ更新 ====================================================
//＠1行更新
void MLCD_MONO::updateArray1(unsigned char ln, const char *data){
	//ln = 行番号 0から_height-1まで
	//画素状態記憶用配列に*dataの内容を上書き
	ln = (ln >= _height)? _height-1 : ln;
	
	//データのサイズが_bytesに満たない場合は何もしない
	if((sizeof(data) / sizeof(data[0])) < _bytes) return;
	
	memcpy(R + (ln * _bytes), data, _bytes);
}
//＠複数行更新
void MLCD_MONO::updateArrayN(unsigned char sn, unsigned char en, const char *data){
	//sn = 開始行, en = 終了行
	//画素状態記憶用配列に*dataの内容を上書き
	sn = (sn >= _height)? _height-1 : sn;
	en = (en >= _height)? _height-1 : en;
	
	//snとenが一緒:1行だけ更新
	if(sn == en){
		memcpy(R + (sn * _bytes), data, _bytes);
		return;
	}
	
	//snとenが逆:ひっくり返す
	if(sn > en){
		sn ^= en;
		en ^= sn;
		sn ^= en;
	}
	
	//snが1でenが_height:全部更新
	if((en - sn) == (_height-1)){
		memcpy(R, data, _height * _bytes);
		return;
	}
	
	//dataのサイズ確認
	unsigned short stidx = sn * _bytes;
	unsigned short len = (en - sn) * _bytes;
	if((sizeof(data) / sizeof(data[0])) < len) return;
	
	//画素状態記憶用配列に*dataの内容を上書き
	memcpy(R + stidx, data, len);
}
//＠全部更新
void MLCD_MONO::updateArrayA(const char *data){
	//画素状態記憶用配列に*dataの内容を上書き
	memcpy(R, data, _height * _bytes);
}

//表示データ書き込み(内部配列を使用する) ============================
//＠1行書き込み(内部配列を使用する)
void MLCD_MONO::writeArray1(unsigned char ln){
	//ln = 行番号 0から_height-1まで
	ln = (ln >= _height)? _height-1 : ln;
	
	//内部配列開始位置
	short idx = ln * _bytes;
	
	//コマンド送出フラグ
	busySPI = true;
	
	CS_HIGH;
	WAIT_1US;
	
	send_mode(SH_MLCD_MONO_WRITE | ((comFlag)? SH_MLCD_MONO_COM : SH_MLCD_MONO_DUMMY)); //モード決定
	send_addr(ln+1); //ゲートライン(行番号 1から_heightまで)選択
	
	//1行分のデータ連続書き込み
	for(short i=0; i<_bytes; i++){
		send_data(R[idx+i]);
	}
	
	//転送待ち(16clockダミーを送る)
	send_data(SH_MLCD_MONO_DUMMY);
	send_data(SH_MLCD_MONO_DUMMY);
	
	WAIT_3US;
	CS_LOW;
	WAIT_1US;
	
	//コマンド送出フラグ
	busySPI = false;
}
//＠複数行書き込み(内部配列を使用する)
void MLCD_MONO::writeArrayN(unsigned char sn, unsigned char en){
	//sn = 開始行, en = 終了行
	sn = (sn >= _height)? _height-1 : sn;
	en = (en >= _height)? _height-1 : en;
	
	//snとenが一緒:1行だけ書き込み
	if(sn == en){
		writeArray1(sn);
		return;
	}
	
	//snとenが逆:ひっくり返す
	if(sn > en){
		sn ^= en;
		en ^= sn;
		sn ^= en;
	}
	
	//内部配列開始位置
	short idx = sn * _bytes;
	int rows = en - sn + 1;
	
	//コマンド送出フラグ
	busySPI = true;
	
	CS_HIGH;
	WAIT_1US;
	
	send_mode(SH_MLCD_MONO_WRITE | ((comFlag)? SH_MLCD_MONO_COM : SH_MLCD_MONO_DUMMY)); //モード決定
	
	//データ連続書き込み
	sn++; //0スタートを1スタートに
	en++; //0スタートを1スタートに
	for(short i=0, j, k; i<rows; i++){
		send_addr(sn + i); //ゲートライン(行番号 1から_heightまで)選択
		//1行分のデータ連続書き込み
		for(j=0, k = i * _bytes + idx; j<_bytes; j++){
			send_data(R[j+k]);
		}
		//転送待ち(8clockダミーを送る)
		send_data(SH_MLCD_MONO_DUMMY);
	}
	//転送待ち(8clockダミーを送る)
	send_data(SH_MLCD_MONO_DUMMY);
	
	WAIT_3US;
	CS_LOW;
	WAIT_1US;
	
	//コマンド送出フラグ
	busySPI = false;
}
//＠全部書き込み(内部配列を使用する)
void MLCD_MONO::writeArrayA(){
	writeArrayN(0, _height-1);
}

//表示データ書き込み(内部配列を使用しない) ==========================
//＠1行書き込み(内部配列を使用しない)
void MLCD_MONO::write1(unsigned char ln, const char *data){
	//データのサイズが_bytesに満たない場合は何もしない
	if((sizeof(data) / sizeof(data[0])) < _bytes) return;
	
	//ln = 行番号 0から_height-1まで
	ln = (ln >= _height)? _height-1 : ln;
	
	//コマンド送出フラグ
	busySPI = true;
	
	CS_HIGH;
	WAIT_1US;
	
	send_mode(SH_MLCD_MONO_WRITE | ((comFlag)? SH_MLCD_MONO_COM : SH_MLCD_MONO_DUMMY)); //モード決定
	send_addr(ln+1); //ゲートライン(行番号 1から_heightまで)選択
	
	//1行分のデータ連続書き込み
	for(short i=0; i<_bytes; i++){
		send_data(data[i]);
	}
	
	//転送待ち(16clockダミーを送る)
	send_data(SH_MLCD_MONO_DUMMY);
	send_data(SH_MLCD_MONO_DUMMY);
	
	WAIT_3US;
	CS_LOW;
	WAIT_1US;
	
	//コマンド送出フラグ
	busySPI = false;
	
}
//＠複数行書き込み(内部配列を使用しない)
void MLCD_MONO::writeN(unsigned char sn, unsigned char en, const char *data){
	//sn = 開始行, en = 終了行
	sn = (sn >= _height)? _height-1 : sn;
	en = (en >= _height)? _height-1 : en;
	
	//snとenが一緒:1行だけ書き込み
	if(sn == en){
		write1(sn, data);
		return;
	}
	
	//snとenが逆:ひっくり返す
	if(sn > en){
		sn ^= en;
		en ^= sn;
		sn ^= en;
	}
	
	//dataのサイズ確認
	unsigned short rows = en - sn + 1;
	unsigned short len = rows * _bytes;
	if((sizeof(data) / sizeof(data[0])) < len) return;
	
	//コマンド送出フラグ
	busySPI = true;
	
	CS_HIGH;
	WAIT_1US;
	
	send_mode(SH_MLCD_MONO_WRITE | ((comFlag)? SH_MLCD_MONO_COM : SH_MLCD_MONO_DUMMY)); //モード決定
	
	//データ連続書き込み
	sn++; //0スタートを1スタートに
	en++; //0スタートを1スタートに
	for(short i=0, j, k; i<rows; i++){
		send_addr(sn + i); //ゲートライン(行番号 1から_heightまで)選択
		//1行分のデータ連続書き込み
		for(j=0, k=i*_bytes; j<_bytes; j++){
			send_data(data[j+k]);
		}
		//転送待ち(8clockダミーを送る)
		send_data(SH_MLCD_MONO_DUMMY);
	}
	//転送待ち(8clockダミーを送る)
	send_data(SH_MLCD_MONO_DUMMY);
	
	WAIT_3US;
	CS_LOW;
	WAIT_1US;
	
	//コマンド送出フラグ
	busySPI = false;
}
//＠全部書き込み(内部配列を使用しない)
void MLCD_MONO::writeA(const char *data){
	writeN(0, _height-1, data);
}

//図形描画 ==========================================================
//＠打点(1ピクセルのドットを打つ)
void MLCD_MONO::pixel(unsigned short x, unsigned short y, char mode, bool immidiate){
	if(x >= _width || y >= _height) return;
	
	short idx = (_bytes * y) + (x >> 3); //内部配列の場所を特定
	char bitidx = 7 - x % 8; //ビットの位置を特定
	
	//内部配列更新
	switch(mode){
		case SH_MLCD_MONO_BLACK: //打点(0)
			R[idx] &= ~(1<<bitidx);
			break;
		case SH_MLCD_MONO_ERASE: //消去(1)
			R[idx] |= 1<<bitidx;
			break;
		case SH_MLCD_MONO_INVERT: //反転(-1)
			R[idx] ^= 1<<bitidx;
			break;
	}
	
	//即時表示の場合は書き込み
	if(immidiate) writeArray1(y);
}
//＠線を描画
void MLCD_MONO::drawLine(unsigned short fromX, unsigned short fromY, unsigned short toX, unsigned short toY, char mode, bool immidiate){
	if(fromX>=_width || toX>=_width || fromY>=_height || toY>=_height) return;
	
	short dx = toX - fromX;
	short dy = toY - fromY;
	dx = (dx < 0)? dx * -1 : dx;
	dy = (dy < 0)? dy * -1 : dy;
	
	//垂直線
	if(dx == 0){
		if(fromY > toY){fromY ^= toY; toY ^= fromY; fromY ^= toY;}//swap
		
		short idx = (_bytes * fromY) + (fromX >> 3); //内部配列の場所を特定
		char bitidx = 7 - fromX % 8; //ビットの位置を特定
		
		//内部配列更新
		char bits = 0;
		switch(mode){
			case SH_MLCD_MONO_BLACK: //打点(0)
				bits = ~(1<<bitidx);
				for(; fromY<=toY; fromY++, idx+=_bytes) R[idx] &= bits;
				break;
			case SH_MLCD_MONO_ERASE: //消去(1)
				bits = 1<<bitidx;
				for(; fromY<=toY; fromY++, idx+=_bytes) R[idx] |= bits;
				break;
			case SH_MLCD_MONO_INVERT: //反転(-1)
				bits = 1<<bitidx;
				for(; fromY<=toY; fromY++, idx+=_bytes) R[idx] ^= bits;
				break;
		}
		
		//即時更新の場合はすぐに表示
		if(immidiate) writeArrayN(toY - dy, toY); //fromY + 1, toY + 1
		
		return;
	}
	
	//水平線
	if(dy == 0){
		if(fromX > toX){fromX ^= toX; toX ^= fromX; fromX ^= toX;}//swap
		
		short idx = (_bytes * fromY) + (fromX >> 3); //内部配列の場所を特定
		
		//ビットマスク作成のためのシフト量とビット位置を算出
		char stbitRef = fromX % 8;
		char stbitIdx = 7 - stbitRef;
		char enbitIdx = toX % 8;
		char enbitRef = 7 - enbitIdx;
		
		//ビットマスク作成
		unsigned char stbits = ((0xFF << stbitRef) & 0xFF) >> stbitRef; //演算はintで実施される
		unsigned char enbits = (0xFF >> enbitRef) << enbitRef;
		
		//塗りつぶす領域が同じ領域内の場合
		if((fromX >> 3) == (toX >> 3)){
			//内部配列更新
			switch(mode){
				case SH_MLCD_MONO_BLACK: //打点(0)
					R[idx] &= ~(stbits & enbits);
					break;
				case SH_MLCD_MONO_ERASE: //消去(1)
					R[idx] |= (stbits & enbits);
					break;
				case SH_MLCD_MONO_INVERT: //反転(-1)
					R[idx] ^= (stbits & enbits);
					break;
			}
		}else{
			//8bit全埋めマスの個数
			char count8 = (dx - (stbitIdx + enbitIdx)) >> 3;
			
			//内部配列更新
			switch(mode){
				case SH_MLCD_MONO_BLACK: //打点(0)
					R[idx] &= ~stbits;
					for(idx++; count8>0; count8--, idx++) R[idx] &= 0x00;
					R[idx] &= ~enbits;
					break;
				case SH_MLCD_MONO_ERASE: //消去(1)
					R[idx] |= stbits;
					for(idx++; count8>0; count8--, idx++) R[idx] |= 0xFF;
					R[idx] |= enbits;
					break;
				case SH_MLCD_MONO_INVERT: //反転(-1)
					R[idx] ^= stbits;
					for(idx++; count8>0; count8--, idx++) R[idx] ^= 0xFF;
					R[idx] ^= enbits;
					break;
			}
		}
		
		//即時更新の場合はすぐに表示
		if(immidiate) writeArray1(fromY);
		
		return;
	}
	
	//傾きのある直線描画：ブレゼンハムのアルゴリズム(Bresenham's line algorithm)
	short sx = (fromX < toX)? 1:-1, sy = (fromY < toY)? 1:-1;
	short err = dx - dy;
	short e2;
	
	for(;;){
		pixel(fromX, fromY, mode, immidiate);
		if((fromX == toX) && (fromY == toY)) break;
		e2 = err * 2;
		if(e2 > -dy){
			err -= dy;
			fromX += sx;
		}
		if(e2 < dx){
			err += dx;
			fromY += sy;
		}
	}
	
}
//＠四角形を描画
void MLCD_MONO::drawRect(unsigned short left, unsigned short top, unsigned short width, unsigned short height, char mode, bool immidiate){
	if(left>_width || top>_height || width<2 || height<2) return;
	
	int toX = left + width;
	int toY = top + height;
	
	if(toX>=_width) toX = _width-1;
	if(toY>=_height) toY = _height-1;
	
	drawLine(left, top, left, toY, mode, immidiate);
	drawLine(left, top, toX, top, mode, immidiate);
	drawLine(left, toY, toX, toY, mode, immidiate);
	drawLine(toX, top, toX, toY, mode, immidiate);
}
//＠対角点指定(左上、右下)での四角形の描画
void MLCD_MONO::drawRect2(unsigned short x1, unsigned short y1, unsigned short x2, unsigned short y2, char mode, bool immidiate){
	if(x1>=_width || y1>=_height || x2>=_width || y2>=_height) return;
	
	if(x1>x2){x1^=x2;x2^=x1;x1^=x2;} //swap
	if(y1>y2){y1^=y2;y2^=y1;y1^=y2;} //swap
	
	if((x2-x1) < 1 || (y2-y1) < 1) return;
	
	drawLine(x1, y1, x1, y2, mode, immidiate);
	drawLine(x1, y1, x2, y1, mode, immidiate);
	drawLine(x1, y2, x2, y2, mode, immidiate);
	drawLine(x2, y1, x2, y2, mode, immidiate);
}
//＠塗りつぶされた四角形を描画
void MLCD_MONO::fillRect(unsigned short left, unsigned short top, unsigned short width, unsigned short height, char mode, bool immidiate){
	if(left>_width || top>_height || width<2 || height<2) return;
	
	int toX = left + width;
	int toY = top + height;
	
	if(toX>=_width) toX = _width-1;
	if(toY>=_height) toY = _height-1;
	
	//水平線を高さ分書き込む
	for(; top<=toY; top++) drawLine(left, top, toX, top, mode, immidiate);
}
//＠対角点指定(左上、右下)での塗りつぶされた四角形を描画
void MLCD_MONO::fill2Point(unsigned short x1, unsigned short y1, unsigned short x2, unsigned short y2, char mode, bool immidiate){
	if(x1>=_width || y1>=_height || x2>=_width || y2>=_height) return;
	
	if(x1>x2){x1^=x2;x2^=x1;x1^=x2;} //swap
	if(y1>y2){y1^=y2;y2^=y1;y1^=y2;} //swap
	
	if((x2-x1) < 1 || (y2-y1) < 1) return;
	
	//水平線を高さ分書き込む
	for(; y1<=y2; y1++) drawLine(x1, y1, x2, y1, mode, immidiate);
}
//＠円の描画
void MLCD_MONO::drawCircle(unsigned short centerX, unsigned short centerY, unsigned short r, char mode, bool immidiate){
	if(centerX>=_width || centerY>=_height || r<1) return;
	
	//中間点円弧描画アルゴリズム(Bresenham's Midpoint circle algorithm)
	int cx=r, cy=0, err=1-cx;
	while(cx>=cy){
		pixel(centerX+cy, centerY-cx, mode, immidiate); //0:12時から45° 時計回り
		pixel(centerX+cx, centerY-cy, mode, immidiate); //1:3時から45° 反時計回り
		pixel(centerX+cx, centerY+cy, mode, immidiate); //2:3時から45° 時計回り
		pixel(centerX+cy, centerY+cx, mode, immidiate); //3:6時から45° 反時計回り
		pixel(centerX-cy, centerY+cx, mode, immidiate); //4:6時から45° 時計回り
		pixel(centerX-cx, centerY+cy, mode, immidiate); //5:9時から45° 反時計回り
		pixel(centerX-cx, centerY-cy, mode, immidiate); //6:9時から45° 時計回り
		pixel(centerX-cy, centerY-cx, mode, immidiate); //7:12時から45° 反時計回り
		cy++;
		if(err<0) err += cy*2 + 1;
		else err += 2*(cy - --cx + 1);
	}
}
//＠塗りつぶされた円の描画
void MLCD_MONO::fillCircle(unsigned short centerX, unsigned short centerY, unsigned short r, char mode, bool immidiate){
	if(centerX>=_width || centerY>=_height || r<1) return;
	
	//中間点円弧描画アルゴリズム(Bresenham's Midpoint circle algorithm)
	int cx=r, cy=0, err=1-cx;
	while(cx>=cy){
		fill2Point(centerX-cy, centerY-cx, centerX+cy, centerY+cx, mode, immidiate);
		fill2Point(centerX-cx, centerY-cy, centerX+cx, centerY+cy, mode, immidiate);
		cy++;
		if(err<0) err += cy*2 + 1;
		else err += 2*(cy - --cx + 1);
	}
}

//画像の描画 ========================================================
//＠画像データ配列を描画
char MLCD_MONO::drawMonoData(const char *data, unsigned short left, unsigned short top, unsigned short imgWidth, unsigned short imgHeight, char mode, bool immidiate){
	if(left>=_width || top>=_height || imgWidth < 2 || imgHeight < 2) return -1;
	if((sizeof(data)/sizeof(data[0])) < ((imgWidth>>3)*imgHeight)) return -1;
	
	/*8bitごとに区切って1行一気に更新しようと考えたが
	処理が非常に煩雑になる上、それほど速度も向上しない
	ためシンプルな方法で更新する。dataを8bitずつ読み取
	ってシフトして1が出たらピクセル描画する。*/
	
	short imgBytes = imgWidth >> 3;
	
	//処理する行数:画面外の処理はしない(0からの繰り返し回数なので-1している)
	short enRows = _height - top - 1;
	enRows = (enRows > imgHeight)? imgHeight - 1 : enRows;
	
	char wk;
	
	for(short hi=0, row=0, hp=top+hi; hi<enRows; hi++, hp++){
		row = hi * imgWidth;
		for(short wi=0, wp=0; wi<imgBytes; wi++){
			wp = left + (wi << 3); //wp = left + (8 * wi);
			if((wp-1) > _width) break; //既に画面外であればそれ以降の処理は省く
			wk = data[row + wi];
			for(char i=0; i<8; i++){
				if((wk<<i) & 0x80) pixel(wp+i, hp, mode, false);
			}
		}
	}
	
	if(immidiate){
		writeArrayN(top, top+enRows);
		return SUCCESS_PRINTED;
	}
	
	return SUCCESS_UPDATE;
}
//＠SDからBMPを読み取って描画
char MLCD_MONO::drawBMPfromSD(char *path, bool BtoW, unsigned short left, unsigned short top, char mode, bool immidiate){
	if(left>=_width || top>=_height) return -1;
	//if(!SD.begin()) return ERR_CANNOT_BEGIN; //SD.begin()がtrueではない
	SD.begin(); //「2回目以降必ずfalse問題」のため判別不可能
	if(!SD.exists(path)) return ERR_NOT_EXISTS; //指定したパスが存在しない
	
	File f = SD.open(path);
	
	if(!f) return ERR_CANNOT_OPEN; //指定したパスは存在するが、開くことができない
	
	//BMPFILEHEADERのbfType(2byte)が'BM'(0x424D)ならBitmapファイル
	//実際はLSBfirstなので'MB'(0x4D42)になる
	if(fr2(f) != 0x4D42){
		f.close();
		return ERR_ISNOT_BITMAP; //対象がBMPファイルではない
	}
	
	fr4(f); //BMPFILEHEADERのbfSize(4byte):無視
	fr4(f); //BMPFILEHEADERのbfReserved1(2byte):無視, bfReserved2(2byte):無視
	
	//BMPFILEHEADERのbfOffBits(4byte):ファイル先頭から画像データまでのオフセット
	int bmpOffset = fr4(f);
	
	//BITMAPINFOHEADERのbiSize(4byte):情報ヘッダサイズが40以下は解析できない
	if(fr4(f) < 40){
		f.close();
		return ERR_IS_OLDBITMAP; //対象が古いOS/2ビットマップなので解析できない
	}
	
	//BITMAPINFOHEADERのbiWidth(4byte)とbiHeight(4byte):画像の幅と高さ
	int bmpWidth = fr4(f);
	int bmpHeight = fr4(f);
	
	//bmpWidth、bmpHeightが負数:特殊なBMPファイルは解析できない
	if(bmpWidth<0 || bmpHeight<0){
		f.close();
		return ERR_IS_SPECIALBITMAP; //対象が特殊なフォーマットなので解析できない
	}
	
	fr2(f); //BITMAPINFOHEADERのbiPlanes(2byte):常に1なので無視
	
	//BITMAPINFOHEADERのbiBitCount(2byte):色深度
	unsigned short bmpDepth = fr2(f);
	
	//色深度16色、256色、65536色はカラーパレットを使うため再現できない
	if(bmpDepth != 1 && bmpDepth < 24){
		f.close();
		return ERR_COLOR_DEPTH; //対象が色深度16色、256色、65536色なので再現できない
	}
	
	//BITMAPINFOHEADERのbiCopmression(4byte):無圧縮のみ対応
	if(fr4(f) != 0){
		f.close();
		return ERR_COMPRESSED; //対象が圧縮されているので解析できない
	}
	
	//BITMAPINFOHEADERの残りの情報は無視
	f.seek(bmpOffset);
	
	//1行あたりのデータサイズ(バイト数)を4の倍数に
	short rs = ceil((bmpWidth * bmpDepth) / 8);
	rs = ((rs % 4)==0)? rs : ((rs >> 2) + 1) << 2;
	
	//処理する行数:画面外の処理はしない
	short enRows = _height - top;
	enRows = (enRows > bmpHeight)? bmpHeight : enRows;
	
	//BMPデータ行先頭アドレス算出用
	unsigned long pos = 0;
	
	//データを取り出していく
	switch(bmpDepth){
		case 1:{//モノクロ2値
			
			char wk = 0;
			
			//bmpは下から記録されているので無理矢理上から処理していく
			//BtoW:Black to White:白黒を逆転させて表示する
			if(BtoW){ //白黒逆転表示
				for(short rows=0, rp=top+rows; rows<enRows; rows++, rp++){
					//ファイルポジションを算出して行の先頭にシーク
					pos = bmpOffset + rs * (bmpHeight - (rows + 1));
					f.seek(pos);
					
					for(short cols=0, cp=0; cols<rs; cols++){
						cp = left + (cols << 3); //cp = left + (8 * cols);
						if((cp-1) > _width) break; //既に画面外であればそれ以降の処理は省く
						wk = f.read();
						for(char i=0; i<8; i++){
							if(!((wk<<i) & 0x80)) pixel(cp+i, rp, mode, false);
						}
					}
				}
			}else{ //正常表示
				for(short rows=0, rp=top+rows; rows<enRows; rows++, rp++){
					//ファイルポジションを算出して行の先頭にシーク
					pos = bmpOffset + rs * (bmpHeight - (rows + 1));
					f.seek(pos);
					
					for(short cols=0, cp=0; cols<rs; cols++){
						cp = left + (cols << 3); //cp = left + (8 * cols);
						if((cp-1) > _width) break; //既に画面外であればそれ以降の処理は省く
						wk = f.read();
						for(char i=0; i<8; i++){
							if((wk<<i) & 0x80) pixel(cp+i, rp, mode, false);
						}
					}
				}
			}
			
			break;
		}
		case 24:{//24bitカラー(フルカラー)
			
			//ディザリング処理(Sierra dithering algorithm - "Sierra Lite")
			short w1 = bmpWidth+1;
			int szw1 = sizeof(float) * w1;
			float r0err[w1];
			float r1err[w1];
			memset(r0err, 0, szw1);
			memset(r1err, 0, szw1);
			if(BtoW){ //白黒逆転表示
				for(short rows=0, rp=top+rows; rows<enRows; rows++, rp++){
					//ファイルポジションを算出して行の先頭にシーク
					pos = bmpOffset + rs * (bmpHeight - (rows + 1));
					f.seek(pos);
					
					for(short cols=0, cp=left+cols; cols<bmpWidth; cols++, cp++){
						if(cp > _width) break; //既に画面外であればそれ以降の処理は省く
						
						//画素の明るさを計算し、0.5以上なら黒
						unsigned short bgr = f.read() + f.read() + f.read();
						float brightness = (bgr / 189.0f) + r0err[cols];
						if(brightness >= 0.5f){
							pixel(cp, rp, mode, false);
						}else{
							brightness -= 1.0f;
						}
						//"Sierra Lite"
						r0err[cols+1] += brightness * 0.5;
						r1err[cols] += brightness * 0.25;
						if(cols > 0) r1err[cols-1] += r1err[cols]; //+= brightness * 0.25;
					}
					
					memcpy(r0err, r1err, szw1);
					memset(r1err, 0, szw1);
				}
			}else{ //正常表示
				for(short rows=0, rp=top+rows; rows<enRows; rows++, rp++){
					//ファイルポジションを算出して行の先頭にシーク
					pos = bmpOffset + rs * (bmpHeight - (rows + 1));
					f.seek(pos);
					
					for(short cols=0, cp=left+cols; cols<bmpWidth; cols++, cp++){
						if(cp > _width) break; //既に画面外であればそれ以降の処理は省く
						
						//画素の明るさを計算し、0.5以上なら白
						unsigned short bgr = f.read() + f.read() + f.read();
						float brightness = (bgr / 765.0f) + r0err[cols];
						if(brightness >= 0.5f){
							brightness -= 1.0f;
						}else{
							pixel(cp, rp, mode, false);
						}
						//"Sierra Lite"
						r0err[cols+1] += brightness * 0.5;
						r1err[cols] += brightness * 0.25;
						if(cols > 0) r1err[cols-1] += r1err[cols]; //+= brightness * 0.25;
					}
					
					memcpy(r0err, r1err, szw1);
					memset(r1err, 0, szw1);
				}
			}
			break;
		}
		case 32:{//32bitカラー(フルカラー)
			
			//ディザリング処理(Sierra dithering algorithm - "Sierra Lite")
			short w1 = bmpWidth+1;
			int szw1 = sizeof(float) * w1;
			float r0err[w1];
			float r1err[w1];
			memset(r0err, 0, szw1);
			memset(r1err, 0, szw1);
			if(BtoW){ //白黒逆転表示
				for(short rows=0, rp=top+rows; rows<enRows; rows++, rp++){
					//ファイルポジションを算出して行の先頭にシーク
					pos = bmpOffset + rs * (bmpHeight - (rows + 1));
					f.seek(pos);
					
					for(short cols=0, cp=left+cols; cols<bmpWidth; cols++, cp++){
						if(cp > _width) break; //既に画面外であればそれ以降の処理は省く
						
						//画素の明るさを計算し、0.5以上なら黒
						unsigned short bgr = f.read() + f.read() + f.read();
						
						//Reservedデータは捨てる
						f.read();
						
						float brightness = (bgr / 765.0f) + r0err[cols];
						if(brightness >= 0.5f){
							pixel(cp, rp, mode, false);
						}else{
							brightness -= 1.0f;
						}
						//"Sierra Lite"
						r0err[cols+1] += brightness * 0.5;
						r1err[cols] += brightness * 0.25;
						if(cols > 0) r1err[cols-1] += r1err[cols]; //+= brightness * 0.25;
					}
					
					memcpy(r0err, r1err, szw1);
					memset(r1err, 0, szw1);
				}
			}else{ //正常表示
				for(short rows=0, rp=top+rows; rows<enRows; rows++, rp++){
					//ファイルポジションを算出して行の先頭にシーク
					pos = bmpOffset + rs * (bmpHeight - (rows + 1));
					f.seek(pos);
					
					for(short cols=0, cp=left+cols; cols<bmpWidth; cols++, cp++){
						if(cp > _width) break; //既に画面外であればそれ以降の処理は省く
						
						//画素の明るさを計算し、0.5以上なら白
						unsigned short bgr = f.read() + f.read() + f.read();
						
						//Reservedデータは捨てる
						f.read();
						
						float brightness = (bgr / 765.0f) + r0err[cols];
						if(brightness >= 0.5f){
							brightness -= 1.0f;
						}else{
							pixel(cp, rp, mode, false);
						}
						//"Sierra Lite"
						r0err[cols+1] += brightness * 0.5;
						r1err[cols] += brightness * 0.25;
						if(cols > 0) r1err[cols-1] += r1err[cols]; //+= brightness * 0.25;
					}
					
					memcpy(r0err, r1err, szw1);
					memset(r1err, 0, szw1);
				}
			}
			break;
		}
	}
	
	//ファイルクローズ
	f.close();
	
	//即時描画
	if(immidiate){
		writeArrayN(top, top+enRows);
		return SUCCESS_PRINTED;
	}
	
	return SUCCESS_UPDATE;
}
//＠SDからBMPを読み取って描画:String
char MLCD_MONO::drawBMPfromSD(String path, bool BtoW, unsigned short left, unsigned short top, char mode, bool immidiate){
	short len = path.length() + 1;
	unsigned char ary[len];
	path.getBytes(ary, len);
	return drawBMPfromSD((char*) ary, BtoW, left, top, mode, immidiate);
}

//文字の描画 ========================================================
//＠一文字を描画
char MLCD_MONO::drawChar(const char *s, unsigned short x, unsigned short y, unsigned char zoom, char mode, bool immidiate){
	if(x>=_width || y>=_height) return -1;
	
	//戻り値と検出インデックス
	char ret = 0;
	unsigned short cidx = 0;
	
	//UTF-8 謎の符号反転への対策
	char one = *s & 0xFF;
	
#ifndef MP12_USE_ZENKAKU //全角文字を使用しない場合
	//フォントのコードの数だけ繰り返す
	for(; cidx<MPF12_TOTAL_CHARS; cidx++){
		//文字コードが合致するものを探す
		if(one == fMPF12[cidx].idx){
			ret = 1;
			break;
		}
	}
#else //全角文字を使用する場合
	if(one<0x20 || one>0x7F){
		//環境依存文字を含まない0x20-0x7Fの領域の外側 = マルチバイト文字の可能性
		for(; cidx<MPJ12R_TOTAL_CHARS; cidx++){
	#if defined MP12_USE_TABLE_UTF8
			if(memcmp(s, &fMPJ12R[cidx].idx[2], 3) == 0){
				ret = 3;
				break;
			}
	#elif defined MP12_USE_TABLE_SJIS
			if(memcmp(s, &fMPJ12R[cidx].idx[0], 2) == 0){
				ret = 2;
				break;
			}
	#endif
		}
	}
	
	//全角文字を探しても見つからなかった場合、半角文字から検索
	if(ret == 0){
		for(cidx = 0; cidx<MPF12_TOTAL_CHARS; cidx++){
			//文字コードが合致するものを探す
			if(one == fMPF12[cidx].idx){
				ret = 1;
				break;
			}
		}
	}
#endif
	
	//文字が見つからなかった ret = 0
	if(ret == 0) return ret;
	
	//描画処理
	if(zoom>1){
		//倍角処理
		short z1 = zoom - 1;
		switch(ret){
			case 1: //半角文字の描画
				for(unsigned char hi=0; hi<MP12_SIZE_H; hi++){
					char g_cpy = fMPF12[cidx].g[hi];
					if(g_cpy == 0) continue;
					for(char wi=0; wi<MP12_SIZE_W_HALF; wi++){
						if(g_cpy<<wi & 0x80){
							short gx = x+wi*zoom, gy = y+hi*zoom;
							fill2Point(gx, gy, gx+z1, gy+z1, mode, immidiate);
						}
					}
				}
				break;
			case 2: //全角文字の描画
			case 3:
				unsigned short g_cpy = 0;
				for(unsigned char hi=0; hi<MP12_SIZE_H; hi++){
					g_cpy = fMPJ12R[cidx].g[hi<<1] << 8 | fMPJ12R[cidx].g[(hi<<1)+1];
					if(g_cpy == 0) continue;
					for(char wi=0; wi<MP12_SIZE_W; wi++){
						if(g_cpy<<wi & 0x8000){
							short gx = x+wi*zoom, gy = y+hi*zoom;
							fill2Point(gx, gy, gx+z1, gy+z1, mode, immidiate);
						}
					}
				}
				break;
		}
	}else{
		switch(ret){
			case 1: //半角文字の描画
				for(unsigned char hi=0; hi<MP12_SIZE_H; hi++){
					char g_cpy = fMPF12[cidx].g[hi];
					if(g_cpy == 0) continue;
					for(char wi=0; wi<MP12_SIZE_W_HALF; wi++){
						if(g_cpy<<wi & 0x80){
							pixel(x+wi, y+hi, mode, immidiate);
						}
					}
				}
				break;
			case 2: //全角文字の描画
			case 3:
				unsigned short g_cpy = 0;
				for(unsigned char hi=0; hi<MP12_SIZE_H; hi++){
					g_cpy = fMPJ12R[cidx].g[hi<<1] << 8 | fMPJ12R[cidx].g[(hi<<1)+1];
					if(g_cpy == 0) continue;
					for(char wi=0; wi<MP12_SIZE_W; wi++){
						if(g_cpy<<wi & 0x8000){
							pixel(x+wi, y+hi, mode, immidiate);
						}
					}
				}
				break;
		}
	}
	
	return ret;
}
//＠倍角指定ありで文字列を描画
char MLCD_MONO::writeString(const char *str, unsigned short x, unsigned short y, unsigned char zoom, char mode, bool immidiate){
	if(x>=_width || y>=_height || zoom<1) return -1;
	
	unsigned short org_x = x;
	
	while(*str){
		//改行処理(\n)
		if(*str == 0x0A){
			x = org_x;
			y += MP12_SIZE_H * zoom;
			str++;
			continue;
		}
		//文字を描画、バイト数を返す
		char charWide = drawChar(str, x, y, zoom, mode, immidiate);
		//バイト数の分だけポインタを進める(半角=1, UTF-8=3, それ以外=2)
		str += (charWide<2)? 1 : charWide;
		x += ((charWide<2)? MP12_SIZE_W_HALF : MP12_SIZE_W) * zoom;
	}
	
	return 1;
}
//＠倍角指定ありで文字列を描画:String
char MLCD_MONO::writeString(const String str, unsigned short x, unsigned short y, unsigned char zoom, char mode, bool immidiate){
	short len = str.length() + 1;
	unsigned char ary[len];
	str.getBytes(ary, len);
	return writeString((char*) ary, x, y, zoom, mode, immidiate);
}
//＠倍角指定なしで文字列を描画
char MLCD_MONO::writeString(const char *str, unsigned short x, unsigned short y, char mode, bool immidiate){
	return writeString(str, x, y, 1, mode, immidiate);
}
//＠倍角指定なしで文字列を描画:String
char MLCD_MONO::writeString(const String str, unsigned short x, unsigned short y, char mode, bool immidiate){
	short len = str.length() + 1;
	unsigned char ary[len];
	str.getBytes(ary, len);
	return writeString((char*) ary, x, y, 1, mode, immidiate);
}
//＠文字列を描画(上記４つの短縮別名)
char MLCD_MONO::ws(const char *str, unsigned short x, unsigned short y, unsigned char zoom, char mode, bool immidiate){
	return writeString(str, x, y, zoom, mode, immidiate);
}
char MLCD_MONO::ws(const String str, unsigned short x, unsigned short y, unsigned char zoom, char mode, bool immidiate){
	return writeString(str, x, y, zoom, mode, immidiate);
}
char MLCD_MONO::ws(const char *str, unsigned short x, unsigned short y, char mode, bool immidiate){
	return writeString(str, x, y, mode, immidiate);
}
char MLCD_MONO::ws(const String str, unsigned short x, unsigned short y, char mode, bool immidiate){
	return writeString(str, x, y, mode, immidiate);
}
