/* GR-SAKURA Sketch Template V2.07 */
#include <Arduino.h>
#include <MsTimer2.h>
#include "SH_MEMORYLCD.h"

//COM反転処理間隔(ミリ秒)
#define TIMER_VAL_MS 200

//モード待ち時間用(ミリ秒)
#define TIMER_VAL_IMGMODE 3500

//MLCD_MONO mlcd(7, 6, 5, 4, 3);
MLCD_MONO mlcd(7, 6, 5);

//COM反転処理
void invCOM(){
    mlcd.invertCOM();
}

char imgMode; //ループでのモード判定用
unsigned long ti; //時間計測用
String modes[] = {"グリッド","図形","文字列","サイン波","画像"};

//セットアップ:最初の1回だけ実行される
void setup(){
	pinMode(PIN_LED1,OUTPUT); //LED1
	pinMode(PIN_LED2,OUTPUT); //LED2
	pinMode(PIN_LED3,OUTPUT); //LED3
	
	Serial.begin(9600);
	
	pinMode(PIN_SW,INPUT); //押しボタン(青)
	
	digitalWrite(PIN_LED1, HIGH); //LED1点灯
	digitalWrite(PIN_LED2, HIGH); //LED2点灯
	digitalWrite(PIN_LED3, HIGH); //LED3点灯
	
	while(digitalRead(PIN_SW) == 1); //押しボタン(青)が押されるまで待ち続ける
	
	digitalWrite(PIN_LED2, LOW); //LED2消灯
	
	mlcd.init(); //液晶初期化
	
	//MsTimer2にinvCOMを登録してタイマー開始
	MsTimer2::set(TIMER_VAL_MS, invCOM);
	MsTimer2::start();
	
	imgMode = 0;
	
	digitalWrite(PIN_LED3, LOW); //LED3消灯
	
	delay(1000);
}

void loop(){
    
    //0から4の中で循環させる
    imgMode = (imgMode + 1) % 5;
    
    //モードに応じて処理を変える
    switch(imgMode){
        //モード0:グリッド
        case 0:{
            //時間計測
            ti = millis();
            
            //垂直線を等間隔に(描画はあとで)
        	for(int x=10; x<SH_MLCD_MONO_PXWIDTH; x+=20){
                mlcd.drawLine(x, 0, x, SH_MLCD_MONO_PXHEIGHT-1, SH_MLCD_MONO_BLACK, false);
        	}
        	
        	//水平線を等間隔に(描画はあとで)
            for(int y=10; y<SH_MLCD_MONO_PXHEIGHT; y+=20){
                mlcd.drawLine(0, y, SH_MLCD_MONO_PXWIDTH-1, y, SH_MLCD_MONO_BLACK, false);
            }
            
            //一気に描画(内部配列を液晶に転送)
            mlcd.writeArrayA();
            
            //時間計測
            ti = millis() - ti;
            
            break;
        }
        //モード1:図形
        case 1:{
            //時間計測
            ti = millis();
            
            //四角形(即時描画)
            mlcd.drawRect(15, 15, 20, 20, SH_MLCD_MONO_BLACK, true);
            mlcd.drawRect2(45, 15, 65, 35, SH_MLCD_MONO_BLACK, true);
            mlcd.fillRect(75, 15, 20, 20, SH_MLCD_MONO_BLACK, true);
            mlcd.fill2Point(105, 15, 125, 35, SH_MLCD_MONO_BLACK, true);
            
            //円(即時描画)
            mlcd.drawCircle(25, 55, 10, SH_MLCD_MONO_BLACK, true);
            mlcd.fillCircle(55, 55, 10, SH_MLCD_MONO_BLACK, true);
            
            //四角形(反転)
            mlcd.fill2Point(19, 19, 95, 95, SH_MLCD_MONO_INVERT, true);
            
            //時間計測
            ti = millis() - ti;
            
            break;
        }
        //モード2:文字列
        case 2:{
            //時間計測
            ti = millis();
            
            //文字列表示(即時描画)
            String str1 = "writeString(:String) zoom x1";
            mlcd.ws(str1, 20, 20);
            
            //文字列表示(即時描画)
            char str2[] = "writeString(:char*) zoom x1";
            mlcd.writeString(str2, 20, 40);
            
            //日本語表示(改行あり)
            str1 = "日本語メッセージも余裕。\nただしメモリをたくさん食います。";
            mlcd.writeString(str1, 20, 60);
            
            //大きい文字
            char str3[] = "Large Text! 大きい文字も余裕。";
            mlcd.ws(str3, 10, 160, 2, SH_MLCD_MONO_BLACK, true);
            
            //時間計測
            ti = millis() - ti;
            
            break;
        }
        //モード3:サイン波
        case 3:{
            //時間計測
            ti = millis();
            
            //3つの120度ずれたサイン波を描画(即時描画)
            float deg120 = 2.09439510239, sH = 65;
            float y1=0, y2=0, y3=0;
            int gW = 200;
            for(float x=100; x<gW; x+=0.15){
                float x0 = x/18.5;
                y1 = sH * sin(x0 - deg120) + sH;
                y2 = sH * sin(x0) + sH;
                y3 = sH * sin(x0 + deg120) + sH;
                mlcd.pixel(x, y1+55);
                mlcd.pixel(x, y2+55);
                mlcd.pixel(x, y3+55);
            }
            
            //時間計測
            ti = millis() - ti;
            
            break;
        }
        //モード4:画像
        case 4:{
            //時間計測
            ti = millis();
            
            //SDカード内のビットマップ画像データ(カラーをディザリング)を表示する
            String fileName = "c.bmp";
            mlcd.drawBMPfromSD(fileName, true, 0, 0);
            
            //時間計測
            ti = millis() - ti;
            
            break;
        }
    }
    
    delay(TIMER_VAL_IMGMODE);
    
    //文字表示部を消す
    mlcd.fill2Point(80, 80, 319, 180, SH_MLCD_MONO_ERASE, true);
    mlcd.drawRect2(80, 80, 319, 180, SH_MLCD_MONO_BLACK, true);
    
    //モード名と処理時間を表示
    mlcd.ws(modes[imgMode], 90, 85);
    char msg[100];
    sprintf(msg, "処理時間 = %ld millis", ti);
    mlcd.ws(msg, 90, 115);
    
    delay(TIMER_VAL_IMGMODE);
    
    //表示を消して、RAMもクリア
    mlcd.cls();
    mlcd.cla();
    
    if(digitalRead(PIN_SW) == 0){
        digitalWrite(PIN_LED1, LOW); //LED1消灯
        exit(1);
    }
}